from __future__ import annotations

from typing import TypeVar

import pydantic
from pydantic import BaseModel


_ModelT = TypeVar('_ModelT', bound=BaseModel)


PYDANTIC_V2 = pydantic.VERSION.startswith('2.')


def model_rebuild(model: type[BaseModel]) -> None:
    if PYDANTIC_V2:
        model.model_rebuild()
    else:
        model.update_forward_refs()  # pyright: ignore[reportDeprecated]


def model_parse_json(model: type[_ModelT], obj: str | bytes) -> _ModelT:
    if PYDANTIC_V2:
        return model.model_validate_json(obj)
    else:
        return model.parse_raw(obj)  # type: ignore



# -*- coding: utf-8 -*-
from __future__ import annotations

import pytest

import json
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from _pytest.pytester import Pytester


@pytest.fixture(autouse=True)
def makeconfig(pytester: Pytester) -> None:
    pytester.makefile('.json', pyrightconfig=json.dumps({'typeCheckingMode': 'strict'}))


def test_collection(pytester: Pytester) -> None:
    pytester.makepyfile(**{'typesafety/example.py': '', 'typesafety/foo/foo.py': ''})

    result = pytester.runpytest('--collect-only')
    outcomes = result.parseoutcomes()
    assert outcomes['tests'] == 2

    result.stdout.fnmatch_lines(
        [
            '*<PyrightTestFile typesafety/example.py>*',
            '*<PyrightTestItem example.py>*',
            '*<PyrightTestFile typesafety/foo/foo.py>*',
            '*<PyrightTestItem foo.py>*',
        ]
    )

    assert result.ret == 0


def test_reveal_type(pytester: Pytester) -> None:
    content = '''
    from typing import Union

    def foo(a: Union[int, str]) -> None:
        if isinstance(a, str):
            reveal_type(a)  # T: str
        else:
            reveal_type(a)  # T: int
    '''
    pytester.makepyfile(**{'typesafety/bar.py': content})
    result = pytester.runpytest()
    result.assert_outcomes(passed=1)


def test_reveal_type_incorrect_comment(pytester: Pytester) -> None:
    content = '''
    def foo(a: str) -> None:
        reveal_type(a)  # T: int
    '''
    pytester.makepyfile(**{'typesafety/bar.py': content})
    result = pytester.runpytest()
    result.assert_outcomes(failed=1)
    result.stdout.fnmatch_lines(
        [
            '1 | def foo(a: str) -> None:',
            '2 |     reveal_type(a)  # T: int',
            'E | Expected revealed type to be "int" but got "str" instead',
        ],
        consecutive=True,
    )


def test_reveal_type_missing_comment(pytester: Pytester) -> None:
    content = '''
    def foo(a: str) -> None:
        reveal_type(a)
    '''
    pytester.makepyfile(**{'typesafety/bar.py': content})
    result = pytester.runpytest()
    result.assert_outcomes(failed=1)
    result.stdout.fnmatch_lines(
        [
            '1 | def foo(a: str) -> None:',
            '2 |     reveal_type(a)',
            'E | Missing type comment, revealed type: str',
        ],
        consecutive=True,
    )


def test_error_message(pytester: Pytester) -> None:
    content = '''
    from typing import Optional

    def foo(a: Optional[str]) -> None:
        print(a.split('.'))  # E: "split" is not a known member of "None"
    '''
    pytester.makepyfile(**{'typesafety/bar.py': content})
    result = pytester.runpytest()
    result.assert_outcomes(passed=1)


def test_error_message_comment_no_error(pytester: Pytester) -> None:
    content = '''
    from typing import Optional

    def foo(a: str) -> None:
        print(a.split('.'))  # E: "split" is not a known member of "None"
    '''
    pytester.makepyfile(**{'typesafety/bar.py': content})
    result = pytester.runpytest()
    result.assert_outcomes(failed=1)
    result.stdout.fnmatch_lines(
        [
            '1 | from typing import Optional',
            '2 | ',
            '3 | def foo(a: str) -> None:',
            '4 |     print(a.split(\'.\'))  # E: "split" is not a known member of "None"',
            'E | Did not raise an error',
        ],
        consecutive=True,
    )
    assert 'pytest_pyright.plugin.PyrightError' not in result.stdout.str()


def test_error_message_missing(pytester: Pytester) -> None:
    content = '''
    from typing import Optional

    def foo(a: Optional[str]) -> None:
        print(a.split('.'))
    '''
    pytester.makepyfile(**{'typesafety/bar.py': content})
    result = pytester.runpytest()
    result.assert_outcomes(failed=1)
    result.stdout.fnmatch_lines(
        [
            '1 | from typing import Optional',
            '2 | ',
            '3 | def foo(a: Optional[str]) -> None:',
            '4 |     print(a.split(\'.\'))',
            'E | Unexpected error: "split" is not a known member of "None"',
        ],
        consecutive=True,
    )
    assert 'pytest_pyright.plugin.PyrightError' not in result.stdout.str()


def test_unexpected_error_first_line(pytester: Pytester) -> None:
    content = '''
    from bad_module import foo
    '''
    pytester.makepyfile(**{'typesafety/bar.py': content})
    result = pytester.runpytest()
    result.assert_outcomes(failed=1)
    result.stdout.fnmatch_lines(
        [
            '1 | from bad_module import foo',
            'E | Unexpected error: Import "bad_module" could not be resolved',
        ],
        consecutive=True,
    )


def test_multiple_errors(pytester: Pytester) -> None:
    content = '''
    from typing import Optional

    def foo(a: Optional[str]) -> None:
        print(a.split('.'))
        if a is not None:
            reveal_type(a)
    '''
    pytester.makepyfile(**{'typesafety/bar.py': content})
    result = pytester.runpytest()
    result.assert_outcomes(failed=1)
    result.stdout.fnmatch_lines(
        [
            '1 | from typing import Optional',
            '2 | ',
            '3 | def foo(a: Optional[str]) -> None:',
            '4 |     print(a.split(\'.\'))',
            'E | Unexpected error: "split" is not a known member of "None"',
            '5 |     if a is not None:',
            '6 |         reveal_type(a)',
            'E | Missing type comment, revealed type: str',
        ],
        consecutive=True,
    )


def test_custom_pyright_directory_commandline(pytester: Pytester) -> None:
    content = '''
    from typing import Optional

    def foo(a: Optional[str]) -> None:
        print(a.split('.'))
    '''
    pytester.makepyfile(**{'my_pyright_tests/bar.py': content})
    result = pytester.runpytest('--collect-only')
    assert result.parseoutcomes() == {}

    result = pytester.runpytest('--pyright-dir=my_pyright_tests', '--collect-only')
    assert result.parseoutcomes() == {'test': 1}


def test_custom_pyright_directory_commandline_multiple_parts(
    pytester: Pytester,
) -> None:
    content = '''
    from typing import Optional

    def foo(a: Optional[str]) -> None:
        print(a.split('.'))
    '''
    pytester.makepyfile(**{'custom_typesafety/pyright/bar.py': content})
    result = pytester.runpytest('--collect-only')
    assert result.parseoutcomes() == {}

    result = pytester.runpytest(
        '--pyright-dir=custom_typesafety/pyright', '--collect-only'
    )
    assert result.parseoutcomes() == {'test': 1}



# -*- coding: utf-8 -*-

import os
import re
import sys
import subprocess
from pathlib import Path
from functools import lru_cache
from typing import Optional, List, Union, Iterator, Any, cast, TYPE_CHECKING

import pytest
import pyright

from _pytest.nodes import Node
from _pytest.config import Config
from _pytest._io import TerminalWriter
from _pytest._code import ExceptionInfo
from _pytest._code.code import TerminalRepr

from .models import PyrightResult, PyrightFile
from ._compat import model_parse_json

if TYPE_CHECKING:
    from _pytest._code.code import _TracebackStyle
    from _pytest.config.argparsing import Parser


# TODO: cleanup code
# TODO: improve performance
# TODO: ini option for custom dir
# TODO: add support for multi-line errors
# TODO: add support for multiple errors on the same line
# TODO: check for unaccessed type information

PYRIGHT_TYPE_RE = re.compile(r'Type of "(.*)" is "(?P<type>.*)"')


def relative_path(path: Path) -> Path:
    return path.relative_to(Path.cwd())


def is_typesafety_file(parent: Node, path: Path) -> bool:
    # TODO: don't know how good this check is
    relative = relative_path(path)
    return relative.as_posix().startswith(parent.config.option.pyright_dir)


def maybe_decode(data: Union[bytes, str]) -> str:
    if isinstance(data, (bytes, bytearray)):
        return data.decode('utf-8')
    if isinstance(data, memoryview):
        raise TypeError(f'Unexpected type: {type(data)}')
    return data


def pytest_collect_file(file_path: Path, parent: Node) -> Optional['PyrightTestFile']:
    if file_path.suffix == '.py' and is_typesafety_file(parent, file_path):
        return PyrightTestFile.from_parent(parent, path=file_path)
    return None


def pytest_addoption(parser: 'Parser'):
    group = parser.getgroup('pyright')
    group.addoption(
        '--pyright-dir',
        action='store',
        default='typesafety',
        help='Specify the root directory to use to search for pyright tests.',
    )


class PyrightTerminalRepr(TerminalRepr):
    def __init__(self, lines: List[str]) -> None:
        self.lines = lines

    def toterminal(self, tw: TerminalWriter) -> None:
        for line in self.lines:
            red = line.startswith('E')
            tw.line(line, bold=red, red=red)

    @classmethod
    def from_error(cls, error: 'PyrightError') -> 'PyrightTerminalRepr':
        return cls(lines=[f'E |  {error.message}'])

    @classmethod
    def from_errors(cls, exc: 'PyrightErrors') -> 'PyrightTerminalRepr':
        """Build a Repr that outputs something like the following

        1 | from typing import Optional
        2 | def foo(a: Optional[str] = None) -> None:
        3 |     a.split('.')
        E | "split" is not a known member of "None"
        """
        # TODO: cleanup
        # TODO: add carets to show where the error ocurred

        def get_separator(lineno: int) -> str:
            return '|'.rjust(max_padding - num_digits(lineno) + 1)

        def num_digits(num: int) -> int:
            # NOTE: this assumes a positive number
            return len(str(num))

        content_lines = exc.item.content.splitlines()
        max_padding = num_digits(len(content_lines)) + 1
        lines = [
            f'{lineno}{get_separator(lineno)} {content}'
            for lineno, content in enumerate(content_lines, start=1)
        ]

        separator = '|'.rjust(max_padding)
        for offset, error in enumerate(
            sorted(exc.errors, key=lambda e: e.lineno), start=0
        ):
            lines.insert(error.lineno + offset, f'E{separator} {error.message}')

        return cls(lines=lines)


class PyrightError(AssertionError):
    def __init__(self, message: str, lineno: int = 0) -> None:
        self.message = message
        self.lineno = lineno
        super().__init__()

    def __str__(self) -> str:
        return self.message


class PyrightErrors(Exception):
    def __init__(self, errors: List[PyrightError], item: 'PyrightTestItem') -> None:
        self.item = item
        self.errors = errors


class PyrightTestItem(pytest.Item):
    def __init__(
        self,
        name: str,
        parent: Optional['PyrightTestFile'] = None,
        config: Optional[Config] = None,
        *,
        path: Path,
    ) -> None:
        super().__init__(name, parent, config)
        self.path = path
        self.starting_lineno = 1

    def runtest(self) -> None:
        file = PyrightFile.parse(self.content)
        process = pyright.run(
            f'--project={self.path.parent}',
            '--outputjson',
            str(self.path),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # https://github.com/microsoft/pyright/blob/main/docs/command-line.md#pyright-exit-codes
        if process.returncode not in {0, 1}:
            print(maybe_decode(process.stderr), file=sys.stderr)
            print(maybe_decode(process.stdout))
            raise PyrightError(
                'An unknown error ocurred while running pyright, '
                'see the captured output for more details.'
            )

        result = model_parse_json(PyrightResult, process.stdout)

        absolute = os.path.normcase(str(self.path.absolute()))
        errors: List[PyrightError] = []

        for diagnostic in result.diagnostics:
            if os.path.normcase(diagnostic.file) != absolute:
                raise PyrightError(
                    f'Received diagnostic for unknown file: {diagnostic.file}; Expected {absolute}'
                )

            # pyright json diagnostic line numbers are 0-based
            line = diagnostic.range.start.line + 1

            if diagnostic.severity == 'error':
                # we only care about the first line
                actual, *_ = diagnostic.message.split('\n')

                try:
                    expected = file.get_error(line)
                except KeyError:
                    errors.append(
                        PyrightError(f'Unexpected error: {actual}', lineno=line)
                    )
                    continue

                if expected != actual:
                    errors.append(
                        PyrightError(
                            f'Expected type error to be \'{expected}\' but '
                            f'got \'{actual}\' instead',
                            lineno=line,
                        )
                    )
                    continue
            elif diagnostic.severity == 'information':
                match = PYRIGHT_TYPE_RE.match(diagnostic.message)
                if match is None:
                    errors.append(
                        PyrightError(
                            f'Could not extract type from message: "{diagnostic.message}"',
                            lineno=line,
                        )
                    )
                    continue

                actual = match.group('type')

                try:
                    expected = file.get_information(line)
                except KeyError:
                    errors.append(
                        PyrightError(
                            f'Missing type comment, revealed type: {actual}',
                            lineno=line,
                        )
                    )
                    continue

                if expected != actual:
                    errors.append(
                        PyrightError(
                            f'Expected revealed type to be "{expected}" but got "{actual}" instead',
                            lineno=line,
                        )
                    )
                    continue
            else:
                errors.append(
                    PyrightError(
                        f'Unknown diagnostic type: {diagnostic.severity}', lineno=line
                    )
                )

        for line, error in file.errors.items():
            if not error.accessed:
                errors.append(PyrightError('Did not raise an error', lineno=line))

        if errors:
            raise PyrightErrors(errors, item=self)

    def repr_failure(
        self,
        excinfo: ExceptionInfo[BaseException],
        style: Optional['_TracebackStyle'] = None,
    ) -> Union[str, TerminalRepr]:
        if isinstance(excinfo.value, PyrightError):
            return PyrightTerminalRepr.from_error(excinfo.value)

        if isinstance(excinfo.value, PyrightErrors):
            return PyrightTerminalRepr.from_errors(excinfo.value)

        return super().repr_failure(excinfo, style=style)

    def reportinfo(self):
        return self.fspath, 0, f'pyright: {self.name}'

    # TODO: does this cache get deleted when the item is destroyed?
    @property
    @lru_cache(maxsize=None)
    def content(self) -> str:
        return self.path.read_text()


class PyrightTestFile(pytest.File):
    @classmethod
    def from_parent(
        cls, *args: Any, **kwargs: Any
    ) -> 'PyrightTestFile':  # pyright: ignore[reportIncompatibleMethodOverride]
        return cast(
            PyrightTestFile,
            super().from_parent(*args, **kwargs),
        )

    def collect(self) -> Iterator[PyrightTestItem]:
        path = Path(self.fspath)
        yield PyrightTestItem.from_parent(parent=self, name=path.name, path=path)



# -*- coding: utf-8 -*-

import re
from typing import List, Dict, Optional

from pydantic import BaseModel, Field

from ._compat import model_rebuild


TYPE_ERROR_RE = re.compile(r'.*# E: (?P<expected>.*)')
REVEAL_TYPE_RE = re.compile(r'\s+reveal_type\(.*\)\s+# T: (?P<expected>.*)')


class Expected(BaseModel):
    message: str
    accessed: bool = False


class PyrightFile(BaseModel):
    errors: Dict[int, Expected]
    informations: Dict[int, Expected]

    @classmethod
    def parse(cls, content: str) -> 'PyrightFile':
        errors: Dict[int, Expected] = {}
        informations: Dict[int, Expected] = {}

        for index, line in enumerate(content.splitlines(), start=1):
            reveal = REVEAL_TYPE_RE.match(line)
            if reveal is not None:
                informations[index] = Expected(message=reveal.group('expected'))
                continue

            error = TYPE_ERROR_RE.match(line)
            if error is not None:
                errors[index] = Expected(message=error.group('expected'))
                continue

        return cls(errors=errors, informations=informations)

    def get_error(self, line: int) -> str:
        error = self.errors[line]
        error.accessed = True
        return error.message

    def get_information(self, line: int) -> str:
        information = self.informations[line]
        information.accessed = True
        return information.message


class PyrightResult(BaseModel):
    time: int
    version: str
    summary: 'PyrightSummary'
    diagnostics: List['PyrightDiagnostic'] = Field(alias='generalDiagnostics')


class PyrightSummary(BaseModel):
    files: int = Field(alias='filesAnalyzed')
    errors: int = Field(alias='errorCount')
    warnings: int = Field(alias='warningCount')
    informations: int = Field(alias='informationCount')
    time_in_seconds: float = Field(alias='timeInSec')


class PyrightDiagnostic(BaseModel):
    file: str
    message: str
    severity: str  # TODO: enum
    range: 'PyrightRange'
    rule: Optional[str] = None


class PyrightRange(BaseModel):
    start: 'PyrightRangeValue'
    end: 'PyrightRangeValue'


class PyrightRangeValue(BaseModel):
    line: int
    character: int


model_rebuild(PyrightRange)
model_rebuild(PyrightResult)
model_rebuild(PyrightDiagnostic)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re
import os
import codecs
from setuptools import setup, find_packages


def read(fname):
    file_path = os.path.join(os.path.dirname(__file__), fname)
    return codecs.open(file_path, encoding='utf-8').read()


version = ''
with open('src/pytest_pyright/__init__.py') as f:
    match = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE)
    if not match:
        raise RuntimeError('version is not set')

    version = match.group(1)

if not version:
    raise RuntimeError('version is not set')


setup(
    name='pytest-pyright',
    version=version,
    author='Robert Craigie',
    author_email='robertcraigie30@gmail.com',
    maintainer='Robert Craigie',
    maintainer_email='robertcraigie30@gmail.com',
    license='MIT',
    url='https://github.com/RobertCraigie/pytest-pyright',
    description='Pytest plugin for type checking code with Pyright',
    long_description=read('README.rst'),
    packages=find_packages(where='src', include=['pytest_pyright', 'pytest_pyright.*']),
    package_dir={'': 'src'},
    python_requires='>=3.7',
    install_requires=read('requirements.txt').splitlines(),
    zip_safe=False,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Pytest',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Testing',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
    ],
    entry_points={
        'pytest11': [
            'pyright = pytest_pyright.plugin',
        ],
    },
)



from syrupy.assertion import SnapshotAssertion
from pytest_pyright.plugin import PyrightErrors, PyrightError, PyrightTerminalRepr


class MockedItem:
    def __init__(self, content: str) -> None:
        self.content = content


def create_formatter(*errors: PyrightError, content: str) -> PyrightTerminalRepr:
    return PyrightTerminalRepr.from_errors(
        PyrightErrors(
            list(errors),
            item=MockedItem(content=content),  # type: ignore
        )
    )


def test_small_file(snapshot: SnapshotAssertion) -> None:
    content = """
    def foo() -> None:
        print(b)
    """
    writer = create_formatter(
        PyrightError('Unexpected error: "b" is not defined', lineno=3), content=content
    )
    assert str(writer) == snapshot


def test_large_file(snapshot: SnapshotAssertion) -> None:
    content = 'def foo() -> None:'
    content += '\n' * 100 + 'print(a)'
    writer = create_formatter(
        PyrightError('Unexpected error: "a" is not defined', lineno=101),
        content=content,
    )
    assert str(writer) == snapshot
